#include "runtime/mem_arena.h"
#include "runtime/resources.h"
#include "runtime/rt_math.h"
#include "runtime/threading.h"

#include "gfx/builtin_objects.h"
#include "gfx/effect.h"
#include "gfx/gfx.h"
#include "gfx/renderer_api.h"

#include "asset_compiler/asset_compiler.h"

#include <stdlib.h>

/* Matches the default vertex layout */

typedef struct {
    rt_v3 pos;
    rt_v3 normal;
    rt_v3 tangent;
    rt_v2 texcoord;
} rt_vertex;

typedef struct {
    const rt_effect *effect;
    rt_buffer_handle vbo;
    uint32_t vertex_count;
} rt_demo_renderable;

static rt_demo_renderable *_renderables;
static unsigned int _renderable_count;

static rt_render_graph *_graph;

static rt_result ForwardPassExecute(uint32_t pass_id,
                                    rt_command_buffer_handle cmdbuf,
                                    const rt_render_view *views,
                                    uint32_t view_count,
                                    void *userdata) {
    RT_ASSERT(view_count == 1, "Expected a single view for the main camera.");
    rt_render_view view = *views;
    RT_VERIFY(rtDoViewTypesMatchExact(view, &g_builtin_render_object_types.render_mesh, 1));

    rt_render_list meshes = view.lists[0];
    if (meshes.length == 0)
        return RT_SUCCESS;

    rt_pipeline_handle bound_pipeline =
        (RT_GET_RENDER_LIST_ELEMENT(meshes, rt_render_mesh, 0)).pipeline;
    g_renderer.CmdBindPipeline(cmdbuf, bound_pipeline);

    for (size_t mesh_idx = 0; mesh_idx < meshes.length; ++mesh_idx) {
        rt_render_mesh mesh = RT_GET_RENDER_LIST_ELEMENT(meshes, rt_render_mesh, mesh_idx);
        if (RT_COMPARE_RENDER_HANDLES(bound_pipeline, mesh.pipeline, !=)) {
            bound_pipeline = mesh.pipeline;
            g_renderer.CmdBindPipeline(cmdbuf, bound_pipeline);
        }
        uint32_t stride = sizeof(rt_vertex);
        g_renderer.CmdBindVertexBuffers(cmdbuf, 0, 1, &mesh.vbo, &stride, NULL);
        g_renderer.CmdDraw(cmdbuf, 0, mesh.vertex_count);
    }

    return RT_SUCCESS;
}

void RegisterCVars(void) {
    rtRegisterAssetCompilerCVars();
}
/* Called after the runtime has finished its initialization and before entering the main-loop*/
void Init(void) {
    rtLog("GAME", "Init");
    rtInitAssetCompiler();

    rtWaitForAssetProcessing();

    rtRegisterBuiltinRenderObjectTypes();

    rt_render_graph_builder builder = g_renderer.CreateRenderGraphBuilder();
    rt_attachment_info backbuffer   = {
          .name    = "backbuffer",
          .format  = RT_PIXEL_FORMAT_SWAPCHAIN,
          .width   = RT_RENDER_TARGET_SIZE_SWAPCHAIN,
          .height  = RT_RENDER_TARGET_SIZE_SWAPCHAIN,
          .samples = 1,
          .layers  = 1,
    };
    builder.AddRenderTarget(builder.obj, &backbuffer);
    rt_attachment_info depth = {.name    = "depth",
                                .format  = RT_PIXEL_FORMAT_DEPTH24_STENCIL8,
                                .width   = RT_RENDER_TARGET_SIZE_SWAPCHAIN,
                                .height  = RT_RENDER_TARGET_SIZE_SWAPCHAIN,
                                .samples = 1,
                                .layers  = 1};
    builder.AddRenderTarget(builder.obj, &depth);

    rt_pass_info forward = {.name  = "forward",
                            .flags = RT_PASS_FLAG_EXECUTE_ALWAYS | RT_PASS_FLAG_GRAPHICS};
    builder.AddRenderPass(builder.obj, &forward);
    builder.AddColorOutput(builder.obj,
                           "forward",
                           "backbuffer",
                           RT_PASS_LOAD_MODE_CLEAR,
                           RT_PASS_WRITE_MODE_STORE,
                           (rt_color){.r = 1.f, .g = 0.f, .b = 1.f, .a = 1.f});
    builder.SetDepthStencilAttachment(builder.obj,
                                      "forward",
                                      "depth",
                                      RT_PASS_LOAD_MODE_CLEAR,
                                      RT_PASS_WRITE_MODE_DISCARD,
                                      (rt_depth_stencil_value){.depth = 1.f, .stencil = 0});
    builder.SetRenderArea(
        builder.obj,
        "forward",
        (rt_rect2){
            .size = {RT_RENDER_TARGET_SIZE_SWAPCHAIN, RT_RENDER_TARGET_SIZE_SWAPCHAIN}
    },
        0.f,
        1.f);
    builder.SetBackbuffer(builder.obj, "backbuffer");
    builder.BindRenderPass(builder.obj, "forward", ForwardPassExecute, NULL);
    if (builder.Build(builder.obj, &_graph) != RT_SUCCESS) {
        rtReportError("GAME", "Failed to build the render graph.");
        return;
    }

    g_renderer.DestroyRenderGraphBuilder(&builder);

    /* Create a simple "scene" */
    _renderable_count = 1;
    _renderables      = calloc(_renderable_count, sizeof(rt_demo_renderable));
    RT_VERIFY(_renderables);
    for (unsigned int i = 0; i < _renderable_count; ++i) {

        rt_vertex tri[3] = {
            {.pos = {0.f, 0.5f, 0.f}},
            {.pos = {-.5f, -.5f, 0.f}},
            {.pos = {.5f, -.5f, 0.f}},
        };

        rt_buffer_handle vbo;
        rt_buffer_info vbo_info = {
            .data  = tri,
            .size  = sizeof(tri),
            .type  = RT_BUFFER_TYPE_VERTEX,
            .usage = RT_BUFFER_USAGE_STATIC,
        };
        if (g_renderer.CreateBuffers(1, &vbo_info, &vbo) != RT_SUCCESS) {
            rtReportError("GAME", "Oh noo...");
            return;
        }
        _renderables[i].vbo          = vbo;
        _renderables[i].vertex_count = 3;

        const rt_effect *effect;
        if (rtLoadEffect(rtGetResourceID("assets/shader/static_object.effect"), &effect) !=
            RT_SUCCESS) {
            rtReportError("GAME", "Oh noo...");
            return;
        }
        _renderables[i].effect = effect;
    }
}

/* Called after exiting the main-loop and before the runtime starts its shutdown */
void Shutdown(void) {
    rtLog("GAME", "Shutdown");

    for (unsigned int i = 0; i < _renderable_count; ++i) {
        rtReleaseEffect(_renderables[i].effect);
        g_renderer.DestroyBuffers(1, &_renderables[i].vbo);
    }

    rtShutdownAssetCompiler();
}

static void DrawRenderables(unsigned int frame_id) {
    rt_create_render_view_result main_view_res =
        rtCreateRenderView(&g_builtin_render_object_types.render_mesh, 1, frame_id);
    if (!main_view_res.ok)
        return;
    rt_render_view main_view   = main_view_res.view;
    const rt_effect_pass *pass = &_renderables[0].effect->passes[0];

    for (unsigned int i = 0; i < _renderable_count; ++i) {
        pass = &_renderables[i].effect->passes[0];
        rt_render_mesh mesh;
        mesh.vbo          = _renderables[i].vbo;
        mesh.vertex_count = _renderables[i].vertex_count;
        mesh.pipeline     = pass->pipeline;
        rtPushRenderObjectToView(&main_view, g_builtin_render_object_types.render_mesh, &mesh);
    }
    rtSubmitRenderView(main_view, _graph, pass->pass_id, frame_id);
}

// Question; How do we move data from update to render.
// This is where we could fill the render views, but that would
// mean double/triple buffering the views
void Update(unsigned int frame_id) {
    DrawRenderables(frame_id);
}

void Render(unsigned int frame_id) {
    g_renderer.ExecuteRenderGraph(_graph, frame_id);
    g_renderer.ResetRenderGraph(_graph, frame_id);
}