#include "effect.h"

#include "runtime/config.h"
#include "runtime/ds.h"
#include "runtime/handles.h"
#include "runtime/hashing.h"
#include "runtime/mem_arena.h"
#include "runtime/threading.h"
#include "runtime/atomics.h"

#include <stdlib.h>
#include <string.h>

RT_CVAR_SZ(rt_EffectCacheSize, "The number of slots in the effect cache. Default: 1024", 1024);

typedef struct {
    rt_resource_id resource;
    rt_effect effect;
    _Alignas(4) unsigned int refcount;
} rt_effect_cache_slot;

/* We use a hashtable to find previously loaded effects.
 * To reclaim unreferenced slots when we need to, we use a minheap.
 * The minheap implements a LRU list. To track usage, we use a global running "usage counter",
 * incremented whenever an effect is loaded.
 */
typedef struct {
    rt_effect_cache_slot *slots;
    rt_hashtable lut;
    rt_minheap reclaim_heap;

    /* Linearly allocate slots until we reach capacity */
    size_t next_free;

    /* Used to track "time" since an effect was loaded */
    _Alignas(4) int usage_counter;

    void *memory;

    rt_rwlock lock;
} rt_effect_cache;

static rt_effect_cache _cache;

rt_result InitEffectCache(void) {
    if (!RT_IS_POWER_OF_TWO(rt_EffectCacheSize.sz)) {
        rtReportError(
            "GFX",
            "The value of \"rt_EffectCacheSize\" must be a power of two.\nConfigured: %zu.",
            rt_EffectCacheSize.sz);
        return RT_INVALID_VALUE;
    }

    rt_create_rwlock_result lock_res = rtCreateRWLock();
    if (!lock_res.ok)
        return RT_UNKNOWN_ERROR;
    _cache.lock = lock_res.lock;

    size_t mem_required = sizeof(rt_effect_cache_slot) * rt_EffectCacheSize.sz +
                          RT_HASH_TABLE_MEMORY_REQUIRED(
                              2 * rt_EffectCacheSize.sz) +      /* double to keep performance up */
                          sizeof(int) * rt_EffectCacheSize.sz + /* heap keys */
                          sizeof(size_t) * rt_EffectCacheSize.sz; /* heap values */
    _cache.memory = malloc(mem_required);
    if (!_cache.memory) {
        rtDestroyRWLock(&_cache.lock);
        return RT_OUT_OF_MEMORY;
    }
    _cache.lut = rtCreateHashtable(rt_EffectCacheSize.sz, _cache.memory, NULL, NULL);

    int *keys =
        (int *)((char *)_cache.memory + RT_HASH_TABLE_MEMORY_REQUIRED(2 * rt_EffectCacheSize.sz));
    size_t *values       = (size_t *)(keys + rt_EffectCacheSize.sz);
    _cache.reclaim_heap  = rtCreateMinheap(keys, values, sizeof(size_t), rt_EffectCacheSize.sz, 0);
    _cache.usage_counter = 0;

    _cache.slots = (rt_effect_cache_slot *)(values + rt_EffectCacheSize.sz);
    memset(_cache.slots, 0, sizeof(rt_effect_cache_slot) * rt_EffectCacheSize.sz);

    return RT_SUCCESS;
}

void ShutdownEffectCache(void) {
    free(_cache.memory);
}

RT_DLLEXPORT uint32_t rtCalculateRenderTargetID(const char *name, size_t len) {
    uint32_t id = rtHashBytes32(name, len);
    if (id == 0)
        id = ~id;
    return id;
}

RT_DLLEXPORT uint32_t rtCalculateRenderPassID(const char *name, size_t len) {
    uint32_t id = rtHashBytes32(name, len);
    if (id == 0)
        id = ~id;
    return id;
}

static void ReleaseEffect(rt_effect *effect) {
    for (unsigned int i = 0; i < effect->pass_count; ++i) {
        g_renderer.DestroyPipeline(effect->passes[i].pipeline);
    }
}

/* Returns the index of the reserved slot */
static size_t ReserveSlot(rt_resource_id id) {
    if (_cache.next_free < rt_EffectCacheSize.sz) {
        size_t slot = _cache.next_free++;
        RT_ASSERT(_cache.slots[slot].refcount == 0, "Got a slot that is still in use.");
        _cache.slots[slot].refcount = 1;

        if (rtHashtableInsert(&_cache.lut, id, slot) != RT_SUCCESS) {
            rtLog("GFX", "Failed to insert effect %x into the lookup table.", id);
            _cache.slots[slot].refcount = 0;
            rtMinheapPush(&_cache.reclaim_heap, _cache.usage_counter, &slot);
            return SIZE_MAX;
        }
        _cache.slots[slot].resource = id;
        return slot;
    } else if (!rtMinheapIsEmpty(&_cache.reclaim_heap)) {
        size_t slot;
        rtMinheapPop(&_cache.reclaim_heap, &slot);
        RT_ASSERT(_cache.slots[slot].refcount == 0, "Got a slot that is still in use.");
        _cache.slots[slot].refcount = 1;

        rt_resource_id old_id       = _cache.slots[slot].resource;
        RT_ASSERT(old_id != RT_INVALID_RESOURCE_ID, "The slot should contain an old effect.");
        ReleaseEffect(&_cache.slots[slot].effect);
        rtHashtableRemove(&_cache.lut, old_id);

        if (rtHashtableInsert(&_cache.lut, id, slot) != RT_SUCCESS) {
            rtLog("GFX", "Failed to insert effect %x into the lookup table.", id);
            _cache.slots[slot].refcount = 0;
            rtMinheapPush(&_cache.reclaim_heap, _cache.usage_counter, &slot);
            return SIZE_MAX;
        }
        _cache.slots[slot].resource = id;
        return slot;
    } else {
        rtLog("GFX",
              "Could not insert effect %x into the cache, because the effect cache is full.",
              id);
        return SIZE_MAX;
    }
}

/* Load resource to memory allocated on the given arena */
static rt_result LoadResource(rt_resource_id id, void **p_out, rt_arena *arena) {
    size_t size = rtGetResourceSize(id);
    if (!size) {
        rtLog("GFX", "ID %x is not a valid resource.", id);
        return RT_INVALID_VALUE;
    }
    void *dst = rtArenaPush(arena, size);
    if (!dst) {
        rtLog("GFX", "Failed to allocate %zu bytes of temporary storage.", size);
        return RT_OUT_OF_MEMORY;
    }
    *p_out = dst;
    return rtGetResource(id, dst);
}

static rt_result LoadEffect(rt_resource_id id, rt_effect *effect) {
    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);
    if (!temp.arena) {
        rtLog("GFX", "Could not get a temporary arena.");
        return RT_OUT_OF_MEMORY;
    }
    const rt_resource *resource = NULL;
    rt_result res               = LoadResource(id, &resource, temp.arena);
    if (res != RT_SUCCESS) {
        rtReturnTemporaryArena(temp);
        return res;
    }
    if (resource->type != RT_RESOURCE_EFFECT) {
        rtReturnTemporaryArena(temp);
        rtLog("GFX", "Resource %x does not refer to an effect resource.", id);
        return RT_INVALID_VALUE;
    }
    const rt_effect_info *effect_info = resource->data;
    effect->pass_count                = effect_info->pass_count;

    for (unsigned int i = 0; i < effect_info->pass_count; ++i) {
        rt_resource *pipeline_resource = NULL;
        res = LoadResource(effect_info->passes[i].pipeline, &pipeline_resource, temp.arena);
        if (res != RT_SUCCESS) {
            rtReturnTemporaryArena(temp);
            return res;
        }
        if (pipeline_resource->type != RT_RESOURCE_PIPELINE) {
            rtReturnTemporaryArena(temp);
            rtLog("GFX", "Resource %x does not refer to a pipeline resource.", id);
            return RT_INVALID_VALUE;
        }
        rt_pipeline_info *pipeline_info = pipeline_resource->data;
        rt_pipeline_handle pipeline     = g_renderer.CompilePipeline(pipeline_info);
        if (!RT_IS_HANDLE_VALID(pipeline)) {
            rtReturnTemporaryArena(temp);
            rtLog("GFX",
                  "Failed to compile the pipeline of pass %d (%x).",
                  i,
                  effect_info->passes[i].pass_id);
            return RT_UNKNOWN_ERROR;
        }
        effect->passes[i].pass_id  = effect_info->passes[i].pass_id;
        effect->passes[i].pipeline = pipeline;
    }

    rtReturnTemporaryArena(temp);

    return RT_SUCCESS;
}

RT_DLLEXPORT rt_result rtLoadEffect(rt_resource_id id, const rt_effect **effect) {
    rtAtomic32Inc(&_cache.usage_counter);

    /* Check if the effect is already loaded */
    rtLockRead(&_cache.lock);
    uint64_t slot = rtHashtableLookup(&_cache.lut, id, UINT64_MAX);
    if (slot != UINT64_MAX) {

        RT_ASSERT(_cache.slots[slot].resource == id, "Got the wrong effect");
        rtAtomic32Inc(&_cache.slots[slot].refcount);
        *effect = &_cache.slots[slot].effect;
        rtUnlockRead(&_cache.lock);
        return RT_SUCCESS;
    }
    rtUnlockRead(&_cache.lock);

    /* Load the effect */
    rtLockWrite(&_cache.lock);
    if (rtHashtableLookup(&_cache.lut, id, UINT64_MAX) != UINT64_MAX) {
        /* Another thread was faster than we, just retry */
        rtUnlockWrite(&_cache.lock);
        return rtLoadEffect(id, effect);
    }
    slot = ReserveSlot(id);
    if (slot == SIZE_MAX) {
        rtUnlockWrite(&_cache.lock);
        return RT_OUT_OF_MEMORY;
    }
    rt_result res = LoadEffect(id, &_cache.slots[slot].effect);
    rtUnlockWrite(&_cache.lock);
    *effect = &_cache.slots[slot].effect;
    return res;
}

RT_DLLEXPORT void rtReleaseEffect(const rt_effect *effect) {
    RT_VERIFY(effect);
    rt_effect_cache_slot *slot = (rt_effect_cache_slot *)((char *)effect - offsetof(rt_effect_cache_slot, effect));
    if (rtAtomic32Dec(&slot->refcount) == 0) {
        rtLockWrite(&_cache.lock);
        size_t slot_index = (size_t)(slot - _cache.slots);
        rtMinheapPush(&_cache.reclaim_heap, _cache.usage_counter, &slot_index);
        rtUnlockWrite(&_cache.lock);
    }
}

