#include "render_list.h"
#include "renderer_api.h"

#include "runtime/config.h"
#include "runtime/mem_arena.h"
#include "runtime/threading.h"

#include <string.h>

RT_CVAR_I(rt_RenderListPoolSize,
          "Size of the pool allocated for render lists in bytes. Default: 8 MiB",
          RT_MB(8));

typedef struct {
    size_t size;
    const char *name;
} rt_render_object_type_data;

typedef struct rt_list_pool_s {
    size_t capacity;
    struct rt_list_pool_s *next;
} rt_list_pool;

typedef struct {
    rt_mutex *lock;
    rt_list_pool *first_free;
    rt_arena arena;

    unsigned int access_frame_id;
} rt_frame_lists;

#define DEFAULT_LIST_CAPACITY RT_KB(1)

static rt_render_object_type_data _types[RT_MAX_RENDER_OBJECT_TYPE + 1];
static unsigned int _type_count = 0;
static rt_rwlock _type_lock;

static rt_frame_lists _frame_lists[4];
static unsigned int _max_frames_in_flight;

rt_result InitRenderLists(void) {
    rt_create_rwlock_result lock_res = rtCreateRWLock();
    if (!lock_res.ok)
        return RT_UNKNOWN_ERROR;
    _type_lock = lock_res.lock;

    _max_frames_in_flight = g_renderer.GetMaxFramesInFlight();
    RT_ASSERT(_max_frames_in_flight > 0 && _max_frames_in_flight < RT_ARRAY_COUNT(_frame_lists),
              "Invalid maxium number of in-flight frames.");

    for (unsigned int i = 0; i < _max_frames_in_flight; ++i) {
        rt_create_arena_result arena_res = rtCreateArena(NULL, (size_t)rt_RenderListPoolSize.i);
        if (!arena_res.ok) {
            rtDestroyRWLock(&_type_lock);
            return RT_OUT_OF_MEMORY;
        }
        _frame_lists[i].arena = arena_res.arena;

        _frame_lists[i].lock = rtCreateMutex();
        if (!_frame_lists[i].lock) {
            rtReleaseArena(&_frame_lists[i].arena);
            rtDestroyRWLock(&_type_lock);
        }

        _frame_lists[i].first_free      = NULL;
        _frame_lists[i].access_frame_id = 0;
    }

    return RT_SUCCESS;
}

void ShutdownRenderLists(void) {
    rtDestroyRWLock(&_type_lock);
    for (unsigned int i = 0; i < _max_frames_in_flight; ++i) {
        rtDestroyMutex(_frame_lists[i].lock);
        rtReleaseArena(&_frame_lists[i].arena);
    }
}

RT_DLLEXPORT rt_render_object_type rtRegisterRenderObjectType(size_t object_size,
                                                              const char *debug_name) {
    if (_type_count == RT_MAX_RENDER_OBJECT_TYPE) {
        rtReportError("GFX", "Too many render object types (max is %u)", RT_MAX_RENDER_OBJECT_TYPE);
        return RT_INVALID_RENDER_OBJECT_TYPE;
    }
    rtLockWrite(&_type_lock);
    rt_render_object_type type = (rt_render_object_type)++_type_count;
    _types[_type_count].size   = object_size;
    _types[_type_count].name   = debug_name;
    if (debug_name)
        rtLog("GFX",
              "Registered render object type %s; object size: %zu. Type: %u",
              debug_name,
              object_size,
              _type_count);
    else
        rtLog("GFX",
              "Registered unnamed render object type; object size: %zu. Type: %u",
              object_size,
              _type_count);
    rtUnlockWrite(&_type_lock);
    return type;
}

RT_DLLEXPORT size_t rtGetRenderObjectSize(rt_render_object_type type) {
    size_t size = 0;
    rtLockRead(&_type_lock);
    if (type > RT_INVALID_RENDER_OBJECT_TYPE && type <= _type_count)
        size = _types[type].size;
    rtUnlockRead(&_type_lock);
    return size;
}

RT_DLLEXPORT const char *rtGetRenderObjectTypeDebugName(rt_render_object_type type) {
    const char *name = NULL;
    rtLockRead(&_type_lock);
    if (type > RT_INVALID_RENDER_OBJECT_TYPE && type <= _type_count)
        name = _types[type].name;
    rtUnlockRead(&_type_lock);
    return name;
}

static rt_create_render_list_result
CreateNewList(rt_render_object_type type, unsigned int frame_id, size_t capacity) {
    rt_create_render_list_result res = {.ok = false};
    unsigned int slot                = frame_id % _max_frames_in_flight;
    rtLockMutex(_frame_lists[slot].lock);

    _frame_lists[slot].access_frame_id = frame_id;

    if (!_frame_lists[slot].first_free ||
        _frame_lists[slot].first_free->capacity < capacity) { /* Allocate a new list */
        rt_list_pool *pool = rtArenaPush(&_frame_lists[slot].arena,
                                         sizeof(rt_list_pool) + sizeof(unsigned int) + capacity);
        if (!pool) {
            rtReportError("GFX",
                          "Out of render list pool space! Configured space: %d kiB",
                          rt_RenderListPoolSize.i / 1024);
            goto out;
        }
        pool->capacity                = capacity;
        pool->next                    = _frame_lists[slot].first_free;
        _frame_lists[slot].first_free = pool;
    }
    rt_render_list list;
    unsigned int *frame_id_store =
        (unsigned int *)((char *)_frame_lists[slot].first_free + sizeof(rt_list_pool));
    *frame_id_store = frame_id;
    list.data = (char *)_frame_lists[slot].first_free + sizeof(rt_list_pool) + sizeof(unsigned int);
    list.type = type;
    list.length                   = 0;
    res.ok                        = true;
    res.list                      = list;
    _frame_lists[slot].first_free = _frame_lists[slot].first_free->next;
out:
    rtUnlockMutex(_frame_lists[slot].lock);
    return res;
}

RT_DLLEXPORT rt_create_render_list_result rtCreateRenderList(rt_render_object_type type,
                                                             unsigned int frame_id) {
    return CreateNewList(type, frame_id, DEFAULT_LIST_CAPACITY);
}

void ResetRenderLists(unsigned int frame_id) {
    unsigned int slot = frame_id % _max_frames_in_flight;
    RT_ASSERT(_frame_lists[slot].access_frame_id == frame_id ||
                  _frame_lists[slot].access_frame_id == 0,
              "Frame id mismatch");
    rtLockMutex(_frame_lists[slot].lock);
    _frame_lists[slot].first_free      = NULL;
    _frame_lists[slot].access_frame_id = 0;
    rtArenaClear(&_frame_lists[slot].arena);
    rtUnlockMutex(_frame_lists[slot].lock);
}

RT_DLLEXPORT bool rtPushRenderListEntry(rt_render_list *list, const void *object) {
    size_t object_size = rtGetRenderObjectSize(list->type);
    rt_list_pool *pool =
        (rt_list_pool *)((char *)list->data - sizeof(rt_list_pool) - sizeof(unsigned int));
    unsigned int frame_id = *(unsigned int *)((char *)list->data - sizeof(unsigned int));
    size_t list_capacity  = pool->capacity / object_size;

    if (list->length == list_capacity) {
        /* "Grow" the list */
        rt_create_render_list_result list_res =
            CreateNewList(list->type, frame_id, pool->capacity * 2);
        if (!list_res.ok)
            return false;

        memcpy(list_res.list.data, list->data, list->length * object_size);

        unsigned int slot = frame_id % _max_frames_in_flight;
        rtLockMutex(_frame_lists[slot].lock);
        pool->next                    = _frame_lists[slot].first_free;
        _frame_lists[slot].first_free = pool;
        rtUnlockMutex(_frame_lists[slot].lock);

        list_res.list.length = list->length;
        *list                = list_res.list;
    }

    char *dst = (char *)list->data + list->length * object_size;
    memcpy(dst, object, object_size);
    ++list->length;
    return true;
}