#include "render_view.h"
#include "renderer_api.h"

#include "runtime/config.h"
#include "runtime/mem_arena.h"
#include "runtime/threading.h"

RT_CVAR_I(rt_RenderViewArenaSize,
          "Size of the memory arena used for allocating render views. Default: 1 MB",
          RT_MB(1));

typedef struct {
    rt_arena arena;
    rt_mutex *lock;
    uint32_t frame_id;
} rt_frame_views;

static rt_frame_views _frames[4];
static unsigned int _max_frames_in_flight;

rt_result InitRenderViews(void) {
    _max_frames_in_flight = g_renderer.GetMaxFramesInFlight();
    RT_ASSERT(_max_frames_in_flight > 0 && _max_frames_in_flight < RT_ARRAY_COUNT(_frames),
              "Invalid maximum number of in-flight frames.");

    for (unsigned int i = 0; i < _max_frames_in_flight; ++i) {
        rt_create_arena_result arena_res = rtCreateArena(NULL, (size_t)rt_RenderViewArenaSize.i);
        if (!arena_res.ok)
            return RT_OUT_OF_MEMORY;
        _frames[i].arena = arena_res.arena;
        _frames[i].lock  = rtCreateMutex();
        if (!_frames[i].lock) {
            rtReleaseArena(&_frames[i].arena);
            return RT_UNKNOWN_ERROR;
        }
        _frames[i].frame_id = 0;
    }
    return RT_SUCCESS;
}

void ShutdownRenderViews(void) {
    for (unsigned int i = 0; i < _max_frames_in_flight; ++i) {
        rtDestroyMutex(_frames[i].lock);
        rtReleaseArena(&_frames[i].arena);
    }
}

void ResetRenderViews(unsigned int frame_id) {
    unsigned int slot = frame_id % _max_frames_in_flight;
    rtArenaClear(&_frames[slot].arena);
}

RT_DLLEXPORT rt_create_render_view_result rtCreateRenderView(const rt_render_object_type *types,
                                                             uint32_t type_count,
                                                             unsigned int frame_id) {
#ifdef RT_DEBUG
    for (uint32_t i = 0; i < type_count - 1; ++i) {
        for (uint32_t j = i + 1; j < type_count; ++j) {
            RT_ASSERT(types[i] != types[j], "Duplicate render list type detected.");
        }
    }
#endif

    unsigned int slot = frame_id % _max_frames_in_flight;

    size_t size = type_count * (sizeof(rt_render_list) + sizeof(rt_render_list));
    rtLockMutex(_frames[slot].lock);
    void *storage = rtArenaPush(&_frames[slot].arena, size);
    _frames[slot].frame_id = frame_id;
    rtUnlockMutex(_frames[slot].lock);
    if (!storage) {
        return (rt_create_render_view_result){
            .ok = false,
        };
    }

    rt_render_view view;
    view.lists      = storage;
    view.list_types = (rt_render_object_type *)(view.lists + type_count);
    view.list_count = type_count;
    view.type_mask  = 0; 

    for (uint32_t i = 0; i < type_count; ++i) {
        rt_create_render_list_result list_res = rtCreateRenderList(types[i], frame_id);
        if (!list_res.ok) {
            return (rt_create_render_view_result){
                .ok = false,
            };
        }
        view.lists[i] = list_res.list;
        view.list_types[i] = types[i];
        view.type_mask |= RT_RENDER_OBJECT_TYPE_BIT(types[i]);
    }

    return (rt_create_render_view_result){.ok = true, .view = view};
}

RT_DLLEXPORT bool
rtPushRenderObjectToView(rt_render_view *view, rt_render_object_type type, const void *object) {
    for (uint32_t i = 0; i < view->list_count; ++i) {
        if (view->list_types[i] == type)
            return rtPushRenderListEntry(&view->lists[i], object);
    }
    return false;
}

RT_DLLEXPORT void
rtSubmitRenderView(rt_render_view view, rt_render_graph *render_graph, uint32_t pass_id, unsigned int frame_id) {
    g_renderer.SubmitRenderView(render_graph, pass_id, view, frame_id);
}

RT_DLLEXPORT int rtDoViewTypesMatchExact(rt_render_view view,
                                         const rt_render_object_type *types,
                                         uint32_t type_count) {
    if (view.list_count != type_count)
        return 0;
    for (uint32_t i = 0; i < type_count; ++i) {
        if (view.list_types[i] != types[i])
            return 0;
    }
    return 1;
}