#ifndef RT_GFX_BACKEND_H
#define RT_GFX_BACKEND_H

/* Backend functions and types. */

#include <stddef.h>

#include "gfx.h"
#include "render_list.h"
#include "render_view.h"

#include "runtime/resources.h"
#include "runtime/rt_math.h"
#include "runtime/runtime.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Handles for backend objects */

#define RT_RENDER_BACKEND_HANDLE_MAX_INDEX   ((1u << 24) - 1)
#define RT_RENDER_BACKEND_HANDLE_MAX_VERSION 255

#define RT_RENDER_BACKEND_HANDLE(name)                                                             \
    typedef struct {                                                                               \
        uint32_t version : 8;                                                                      \
        uint32_t index : 24;                                                                       \
    } name

RT_RENDER_BACKEND_HANDLE(rt_pipeline_handle);
RT_RENDER_BACKEND_HANDLE(rt_render_target_handle);
RT_RENDER_BACKEND_HANDLE(rt_command_buffer_handle);
RT_RENDER_BACKEND_HANDLE(rt_gpu_semaphore_handle);
RT_RENDER_BACKEND_HANDLE(rt_buffer_handle);

#undef RT_RENDER_BACKEND_HANDLE

#define RT_COMPARE_RENDER_HANDLES(_A, _B, _Comp) ((*(uint32_t *)&(_A)) _Comp (*(uint32_t *)&(_B)))

/* Init data for the renderer */

#ifdef _WIN32
struct HINSTANCE__;
struct HWND__;
#elif defined(RT_USE_XLIB)
struct _XDisplay;
#endif

struct rt_renderer_init_info_s {
#ifdef _WIN32
    struct HINSTANCE__ *hInstance;
    struct HWND__ *hWnd;
#elif defined(RT_USE_XLIB)
    struct _XDisplay *display;
    unsigned long window;
#endif
};

/* Argument types for render commands */

typedef enum {
    RT_GRAPHICS_QUEUE,
    RT_COMPUTE_QUEUE,
    RT_TRANSFER_QUEUE,
} rt_gpu_queue;

#if 0
/* Attributes are used to bind buffers (or textures) to symbolic values.
 * For example, an attribute might be bound to "CELL_GRID", which would be
 * replaced with the (at the time of the invoke) grid buffer of the current
 * world cell.
 */

typedef enum {
    RT_ATTRIBUTE_VALUE_UNDEFINED,

    RT_ATTRIBUTE_VALUE_MATERIAL_ALBEDO,
    RT_ATTRIBUTE_VALUE_MATERIAL_NORMAL,

    RT_ATTRIBUTE_VALUE_count
} rt_attribute_value;

typedef struct {
    uint32_t index;
    rt_attribute_value value;
} rt_attribute_binding;

#endif

typedef enum {
    RT_SHADER_TYPE_INVALID,
    RT_SHADER_TYPE_VULKAN,
    RT_SHADER_TYPE_DX11,

    RT_SHADER_TYPE_count,
} rt_shader_type;

typedef enum {
    RT_SHADER_STAGE_VERTEX,
    RT_SHADER_STAGE_FRAGMENT,
    RT_SHADER_STAGE_COMPUTE,

    RT_SHADER_STAGE_count,
} rt_shader_stage;

typedef struct {
    rt_shader_type type;
    rt_shader_stage stage;
    rt_relptr bytecode;
    size_t bytecode_length;
} rt_shader_info;

typedef struct {
    rt_gpu_queue target_queue;
} rt_alloc_command_buffer_info;

typedef struct {
    const rt_command_buffer_handle *command_buffers;
    const rt_gpu_semaphore_handle *wait_semaphores;
    const uint64_t *wait_values;
    const rt_gpu_semaphore_handle *signal_semaphores;
    const uint64_t *signal_values;
    uint32_t command_buffer_count;
    uint32_t wait_semaphore_count;
    uint32_t signal_semaphore_count;
} rt_submit_command_buffers_info;

typedef struct {
    /* Optional, for debug purposes */
    const char *name;
    uint64_t initial_value;
} rt_gpu_semaphore_info;

typedef enum {
    RT_BUFFER_TYPE_VERTEX,
    RT_BUFFER_TYPE_INDEX,
    RT_BUFFER_TYPE_UNIFORM,
    RT_BUFFER_TYPE_STORAGE,
    RT_BUFFER_TYPE_count
} rt_buffer_type;

typedef enum {
    /* Create once, never change the data. */
    RT_BUFFER_USAGE_STATIC,

    /* Update occasionally (after a number of frames) */
    RT_BUFFER_USAGE_DYNAMIC,

    /* Create, use once and then discard */
    RT_BUFFER_USAGE_TRANSIENT,

    RT_BUFFER_USAGE_count,
} rt_buffer_usage;

typedef struct {
    size_t size;
    rt_buffer_type type;
    rt_buffer_usage usage;
    const void *data;
} rt_buffer_info;

typedef enum {
    RT_PASS_LOAD_MODE_LOAD,
    RT_PASS_LOAD_MODE_CLEAR,
} rt_pass_load_mode;

typedef enum {
    RT_PASS_WRITE_MODE_STORE,
    RT_PASS_WRITE_MODE_DISCARD,
} rt_pass_write_mode;

typedef union {
    rt_color color;
    struct {
        float depth;
        int32_t stencil;
    } depth_stencil;
} rt_pass_clear_value;

typedef struct {
    float depth;
    int32_t stencil;
} rt_depth_stencil_value;

typedef struct {
    rt_render_target_handle color_buffers[4];
    rt_pass_load_mode color_buffer_loads[4];
    rt_pass_write_mode color_buffer_writes[4];
    rt_pass_clear_value color_buffer_clear_values[4];
    uint32_t color_buffer_count;

    rt_render_target_handle depth_stencil_buffer;
    rt_pass_load_mode depth_stencil_buffer_load;
    rt_pass_write_mode depth_stencil_buffer_write;
    rt_pass_clear_value depth_stencil_buffer_clear_value;

    rt_rect2i render_area;

    // For debug purposes, can be NULL
    const char *name;
} rt_cmd_begin_pass_info;

typedef enum {
    /* Unusable, must be transitioned to an usable state first. */
    RT_RENDER_TARGET_STATE_INVALID,

    /* Used as a color- or depth-buffer */
    RT_RENDER_TARGET_STATE_ATTACHMENT,

    RT_RENDER_TARGET_STATE_SAMPLED_IMAGE,

    RT_RENDER_TARGET_STATE_STORAGE_IMAGE,
} rt_render_target_state;

#define RT_RENDER_TARGET_SIZE_SWAPCHAIN 0

/* Renderer API */

typedef struct rt_pipeline_info_s rt_pipeline_info;

typedef struct {
    const char *name;
    rt_pixel_format format;
    unsigned int width;
    unsigned int height;

    unsigned int samples;
    unsigned int layers;
} rt_attachment_info;

enum {
    /* Bit 0 contains the type: 0 -> graphics, 1 -> compute */
    RT_PASS_FLAG_GRAPHICS  = 0x0000,
    RT_PASS_FLAG_COMPUTE   = 0x0001,
    RT_PASS_FLAG_TYPE_MASK = RT_PASS_FLAG_COMPUTE | RT_PASS_FLAG_GRAPHICS,

    /* Always excecute the pass, even if no objects will be rendered. */
    RT_PASS_FLAG_EXECUTE_ALWAYS = 0x0002,
};

typedef struct {
    const char *name;
    uint32_t flags;
} rt_pass_info;

typedef struct rt_render_graph_s rt_render_graph;
typedef rt_result rt_execute_render_pass_fn(uint32_t pass_id,
                                            rt_command_buffer_handle cmdbuf,
                                            const rt_render_view *views,
                                            unsigned int view_count,
                                            void *userdata);

typedef struct {
    void *obj;

    void (*AddRenderTarget)(void *obj, const rt_attachment_info *info);

    void (*SetBackbuffer)(void *obj, const char *rt_name);

    void (*AddRenderPass)(void *obj, const rt_pass_info *info);
    void (*AddColorOutput)(void *obj,
                           const char *pass_name,
                           const char *rt_name,
                           rt_pass_load_mode load,
                           rt_pass_write_mode write,
                           rt_color clear_color);
    void (*AddSampledInput)(void *obj, const char *pass_name, const char *rt_name);
    void (*SetDepthStencilAttachment)(void *obj,
                                      const char *pass_name,
                                      const char *rt_name,
                                      rt_pass_load_mode load,
                                      rt_pass_write_mode write,
                                      rt_depth_stencil_value clear_value);
    void (*SetRenderArea)(void *obj, const char *pass_name, rt_rect2 area, float min_depth, float max_depth);
    void (*BindRenderPass)(void *obj,
                           const char *pass_name,
                           rt_execute_render_pass_fn *execute_fn,
                           void *userdata);

    rt_result (*Build)(void *obj, rt_render_graph **p_render_graph);
} rt_render_graph_builder;

typedef void rt_register_renderer_cvars_fn(void);
typedef rt_result rt_init_renderer_fn(const rt_renderer_init_info *info);
typedef void rt_shutdown_renderer_fn(void);
typedef unsigned int rt_get_max_frames_in_flight_fn(void);
typedef void rt_begin_frame_fn(unsigned int frame_id);
typedef void rt_end_frame_fn(unsigned int frame_id);
typedef rt_pipeline_handle rt_compile_pipeline_fn(const rt_pipeline_info *info);
typedef void rt_destroy_pipeline_fn(rt_pipeline_handle handle);
typedef rt_result rt_alloc_command_buffers_fn(uint32_t count,
                                              const rt_alloc_command_buffer_info *info,
                                              rt_command_buffer_handle *p_command_buffers);
typedef rt_result rt_submit_command_buffers_fn(rt_gpu_queue queue,
                                               const rt_submit_command_buffers_info *info);
typedef rt_result
rt_create_buffers_fn(uint32_t count, const rt_buffer_info *info, rt_buffer_handle *p_buffers);
typedef void rt_destroy_buffers_fn(uint32_t count, rt_buffer_handle *buffers);

typedef rt_render_graph_builder rt_create_render_graph_builder_fn(void);
typedef void rt_destroy_render_graph_builder_fn(rt_render_graph_builder *builder);
typedef rt_result rt_execute_render_graph_fn(rt_render_graph *rgraph, unsigned int frame_id);
typedef void
rt_submit_render_view_fn(rt_render_graph *render_graph, uint32_t pass_id, rt_render_view view, unsigned int frame_id);
typedef void rt_reset_render_graph_fn(rt_render_graph *graph, unsigned int frame_id);

typedef void rt_cmd_begin_pass_fn(rt_command_buffer_handle cmdbuf,
                                  const rt_cmd_begin_pass_info *info);
typedef void rt_cmd_end_pass_fn(rt_command_buffer_handle cmdbuf);
typedef void rt_cmd_transition_render_target_fn(rt_command_buffer_handle cmdbuf,
                                                rt_render_target_handle render_target,
                                                rt_render_target_state new_state);
typedef void rt_cmd_flush_render_target_write_fn(rt_command_buffer_handle cmdbuf,
                                                 rt_render_target_handle render_target);
typedef void rt_cmd_bind_pipeline_fn(rt_command_buffer_handle cmd, rt_pipeline_handle pipeline);
typedef void rt_cmd_bind_vertex_buffers_fn(rt_command_buffer_handle cmd,
                                           uint32_t first_binding,
                                           uint32_t count,
                                           const rt_buffer_handle *buffers,
    const uint32_t *strides,
                                           const uint32_t *offsets);
typedef void
rt_cmd_draw_fn(rt_command_buffer_handle cmdbuf, uint32_t first_vertex, uint32_t vertex_count);

typedef struct {
    rt_register_renderer_cvars_fn *RegisterCVars;
    rt_init_renderer_fn *Init;
    rt_shutdown_renderer_fn *Shutdown;
    rt_get_max_frames_in_flight_fn *GetMaxFramesInFlight;
    rt_begin_frame_fn *BeginFrame;
    rt_end_frame_fn *EndFrame;
    rt_compile_pipeline_fn *CompilePipeline;
    rt_destroy_pipeline_fn *DestroyPipeline;
    rt_alloc_command_buffers_fn *AllocCommandBuffers;
    rt_submit_command_buffers_fn *SubmitCommandBuffers;
    rt_create_buffers_fn *CreateBuffers;
    rt_destroy_buffers_fn *DestroyBuffers;

    /*render graph functions*/
    rt_create_render_graph_builder_fn *CreateRenderGraphBuilder;
    rt_destroy_render_graph_builder_fn *DestroyRenderGraphBuilder;
    rt_execute_render_graph_fn *ExecuteRenderGraph;
    rt_submit_render_view_fn *SubmitRenderView;
    rt_reset_render_graph_fn *ResetRenderGraph;

    /* Command Buffer Functions */
    rt_cmd_begin_pass_fn *CmdBeginPass;
    rt_cmd_end_pass_fn *CmdEndPass;
    rt_cmd_transition_render_target_fn *CmdTransitionRenderTarget;
    rt_cmd_flush_render_target_write_fn *CmdFlushRenderTargetWrite;
    rt_cmd_bind_pipeline_fn *CmdBindPipeline;
    rt_cmd_bind_vertex_buffers_fn *CmdBindVertexBuffers;
    rt_cmd_draw_fn *CmdDraw;
} rt_renderer_api;

#define RT_RENDERER_API_FN(name) RT_DLLEXPORT rtRen##name

#ifndef RT_DONT_DEFINE_RENDERER_GLOBAL
extern rt_renderer_api g_renderer;
#endif

#ifdef __cplusplus
}
#endif

#endif
